class LocalizedFile(object):
    """ A helper for localized files.

    Automatically choses the file according to the currently used locale. The
    files are internally stored as normal files using the filename to identify
    the wanted file.

    Example:

        class MyModel(Base, AssociatedFiles):
            pdf = LocalizedFile()

    """

    def __set_name__(self, owner, name):
        self.name = name

    def __get_localized_name__(self, instance, locale=None):
        return '{}-{}'.format(
            self.name,
            locale or instance.session_manager.current_locale
        )

    def __get_by_locale__(self, instance, locale=None):
        if instance:
            name = self.__get_localized_name__(instance, locale)
            for file in instance.files:
                if file.name == name:
                    return file

    def __get__(self, instance, owner):
        return self.__get_by_locale__(instance)

    def __set_by_locale__(self, instance, value, locale=None):
        if instance:
            value.name = self.__get_localized_name__(instance, locale)
            self.__delete_by_locale__(instance, locale)
            instance.files.append(value)

    def __set__(self, instance, value):
        return self.__set_by_locale__(instance, value)

    def __delete_by_locale__(self, instance, locale=None):
        if instance:
            name = self.__get_localized_name__(instance, locale)
            for file in instance.files:
                if file.name == name:
                    instance.files.remove(file)

    def __delete__(self, instance):
        self.__delete_by_locale__(instance)



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.swissvotes import _
from onegov.swissvotes.layouts.default import DefaultLayout


class VoteLayout(DefaultLayout):

    @cached_property
    def title(self):
        return self.model.title

    @cached_property
    def editbar_links(self):
        result = []
        if self.request.has_role('admin', 'editor'):
            result.append(
                Link(
                    text=_("Manage attachments"),
                    url=self.request.link(self.model, name='upload'),
                    attrs={'class': 'upload-icon'}
                )
            )
            result.append(
                Link(
                    text=_("Delete vote"),
                    url=self.request.link(self.model, name='delete'),
                    attrs={'class': 'delete-icon'}
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.title, '#'),
        ]


class VoteStrengthsLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Voter strengths")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.model.title, self.request.link(self.model)),
            Link(self.title, '#'),
        ]


class UploadVoteAttachemtsLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Manage attachments")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.model.title, self.request.link(self.model)),
            Link(self.title, '#'),
        ]


class DeleteVoteLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Delete vote")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.model.title, self.request.link(self.model)),
            Link(self.title, '#'),
        ]



from datetime import date
from datetime import datetime
from decimal import Decimal
from freezegun import freeze_time
from onegov.swissvotes.collections import SwissVoteCollection
from psycopg2.extras import NumericRange
from pytz import utc


def test_app(swissvotes_app):
    assert swissvotes_app.principal
    assert swissvotes_app.static_content_pages == {
        'home', 'disclaimer', 'imprint', 'data-protection'
    }


def test_app_dataset_caches(swissvotes_app):
    votes = SwissVoteCollection(swissvotes_app.session())

    csv = [swissvotes_app.get_cached_dataset('csv')]
    xlsx = [swissvotes_app.get_cached_dataset('xlsx')]

    with freeze_time(datetime(2019, 1, 1, 10, tzinfo=utc)):
        votes.add(
            id=1,
            bfs_number=Decimal('100.1'),
            date=date(1990, 6, 2),
            decade=NumericRange(1990, 1999),
            legislation_number=4,
            legislation_decade=NumericRange(1990, 1994),
            title="Vote",
            votes_on_same_day=2,
            _legal_form=1
        )

    csv.append(swissvotes_app.get_cached_dataset('csv'))
    xlsx.append(swissvotes_app.get_cached_dataset('xlsx'))

    with freeze_time(datetime(2019, 7, 9, 12, tzinfo=utc)):
        votes.add(
            id=2,
            bfs_number=Decimal('100.2'),
            date=date(1990, 6, 2),
            decade=NumericRange(1990, 1999),
            legislation_number=4,
            legislation_decade=NumericRange(1990, 1994),
            title="Vote",
            votes_on_same_day=2,
            _legal_form=1
        )

    csv.append(swissvotes_app.get_cached_dataset('csv'))
    xlsx.append(swissvotes_app.get_cached_dataset('xlsx'))

    assert len(csv) == len(set(csv))
    assert len(xlsx) == len(set(xlsx))
    assert set(swissvotes_app.filestorage.listdir('.')) == {
        'dataset-.csv',
        'dataset-.xlsx',
        'dataset-1546336800.0.csv',
        'dataset-1546336800.0.xlsx',
        'dataset-1562673600.0.csv',
        'dataset-1562673600.0.xlsx',
    }



from onegov.core.security import Public
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.swissvotes.models import Principal


@SwissvotesApp.html(
    model=Principal,
    template='home.pt',
    permission=Public
)
def view_home(self, request):
    """ The home page. """

    page = TranslatablePageCollection(request.session).setdefault('home')
    return request.redirect(request.link(page))



from collections import OrderedDict
from dateutil.parser import parse
from decimal import Decimal
from onegov.form.fields import UploadField
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.swissvotes import _
from onegov.swissvotes.models import SwissVote
from onegov.swissvotes.utils import get_table_column
from psycopg2.extras import NumericRange
from xlrd import open_workbook
from xlrd import XL_CELL_EMPTY
from xlrd import xldate


COLUMNS = OrderedDict((
    ('bfs_number', 'anr'),
    ('date', 'datum'),
    ('legislation_number', 'legislatur'),
    ('legislation_decade', 'legisjahr'),
    ('decade', 'jahrzehnt'),
    ('title', 'titel'),
    ('keyword', 'stichwort'),
    ('votes_on_same_day', 'anzahl'),
    ('_legal_form', 'rechtsform'),
    ('descriptor_1_level_1', 'd1e1'),
    ('descriptor_1_level_2', 'd1e2'),
    ('descriptor_1_level_3', 'd1e3'),
    ('descriptor_2_level_1', 'd2e1'),
    ('descriptor_2_level_2', 'd2e2'),
    ('descriptor_2_level_3', 'd2e3'),
    ('descriptor_3_level_1', 'd3e1'),
    ('descriptor_3_level_2', 'd3e2'),
    ('descriptor_3_level_3', 'd3e3'),
    ('_result_people_accepted', 'volk'),
    ('_result_cantons_accepted', 'stand'),
    ('_result', 'annahme'),
    ('result_eligible_voters', 'berecht'),
    ('result_votes_total', 'stimmen'),
    ('result_turnout', 'bet'),
    ('result_votes_empty', 'leer'),
    ('result_votes_invalid', 'ungultig'),
    ('result_votes_valid', 'gultig'),
    ('result_people_yeas', 'volkja'),
    ('result_people_nays', 'volknein'),
    ('result_people_yeas_p', 'volkja-proz'),
    ('result_cantons_yeas', 'kt-ja'),
    ('result_cantons_nays', 'kt-nein'),
    ('result_cantons_yeas_p', 'ktjaproz'),
    ('result_zh_eligible_voters', 'zh-berecht'),
    ('result_zh_votes_total', 'zh-stimmen'),
    ('result_zh_turnout', 'zh-bet'),
    ('result_zh_votes_valid', 'zh-gultig'),
    ('result_zh_yeas', 'zh-ja'),
    ('result_zh_nays', 'zh-nein'),
    ('result_zh_yeas_p', 'zh-japroz'),
    ('_result_zh_accepted', 'zh-annahme'),
    ('result_be_eligible_voters', 'be-berecht'),
    ('result_be_votes_total', 'be-stimmen'),
    ('result_be_turnout', 'be-bet'),
    ('result_be_votes_valid', 'be-gultig'),
    ('result_be_yeas', 'be-ja'),
    ('result_be_nays', 'be-nein'),
    ('result_be_yeas_p', 'be-japroz'),
    ('_result_be_accepted', 'be-annahme'),
    ('result_lu_eligible_voters', 'lu-berecht'),
    ('result_lu_votes_total', 'lu-stimmen'),
    ('result_lu_turnout', 'lu-bet'),
    ('result_lu_votes_valid', 'lu-gultig'),
    ('result_lu_yeas', 'lu-ja'),
    ('result_lu_nays', 'lu-nein'),
    ('result_lu_yeas_p', 'lu-japroz'),
    ('_result_lu_accepted', 'lu-annahme'),
    ('result_ur_eligible_voters', 'ur-berecht'),
    ('result_ur_votes_total', 'ur-stimmen'),
    ('result_ur_turnout', 'ur-bet'),
    ('result_ur_votes_valid', 'ur-gultig'),
    ('result_ur_yeas', 'ur-ja'),
    ('result_ur_nays', 'ur-nein'),
    ('result_ur_yeas_p', 'ur-japroz'),
    ('_result_ur_accepted', 'ur-annahme'),
    ('result_sz_eligible_voters', 'sz-berecht'),
    ('result_sz_votes_total', 'sz-stimmen'),
    ('result_sz_turnout', 'sz-bet'),
    ('result_sz_votes_valid', 'sz-gultig'),
    ('result_sz_yeas', 'sz-ja'),
    ('result_sz_nays', 'sz-nein'),
    ('result_sz_yeas_p', 'sz-japroz'),
    ('_result_sz_accepted', 'sz-annahme'),
    ('result_ow_eligible_voters', 'ow-berecht'),
    ('result_ow_votes_total', 'ow-stimmen'),
    ('result_ow_turnout', 'ow-bet'),
    ('result_ow_votes_valid', 'ow-gultig'),
    ('result_ow_yeas', 'ow-ja'),
    ('result_ow_nays', 'ow-nein'),
    ('result_ow_yeas_p', 'ow-japroz'),
    ('_result_ow_accepted', 'ow-annahme'),
    ('result_nw_eligible_voters', 'nw-berecht'),
    ('result_nw_votes_total', 'nw-stimmen'),
    ('result_nw_turnout', 'nw-bet'),
    ('result_nw_votes_valid', 'nw-gultig'),
    ('result_nw_yeas', 'nw-ja'),
    ('result_nw_nays', 'nw-nein'),
    ('result_nw_yeas_p', 'nw-japroz'),
    ('_result_nw_accepted', 'nw-annahme'),
    ('result_gl_eligible_voters', 'gl-berecht'),
    ('result_gl_votes_total', 'gl-stimmen'),
    ('result_gl_turnout', 'gl-bet'),
    ('result_gl_votes_valid', 'gl-gultig'),
    ('result_gl_yeas', 'gl-ja'),
    ('result_gl_nays', 'gl-nein'),
    ('result_gl_yeas_p', 'gl-japroz'),
    ('_result_gl_accepted', 'gl-annahme'),
    ('result_zg_eligible_voters', 'zg-berecht'),
    ('result_zg_votes_total', 'zg-stimmen'),
    ('result_zg_turnout', 'zg-bet'),
    ('result_zg_votes_valid', 'zg-gultig'),
    ('result_zg_yeas', 'zg-ja'),
    ('result_zg_nays', 'zg-nein'),
    ('result_zg_yeas_p', 'zg-japroz'),
    ('_result_zg_accepted', 'zg-annahme'),
    ('result_fr_eligible_voters', 'fr-berecht'),
    ('result_fr_votes_total', 'fr-stimmen'),
    ('result_fr_turnout', 'fr-bet'),
    ('result_fr_votes_valid', 'fr-gultig'),
    ('result_fr_yeas', 'fr-ja'),
    ('result_fr_nays', 'fr-nein'),
    ('result_fr_yeas_p', 'fr-japroz'),
    ('_result_fr_accepted', 'fr-annahme'),
    ('result_so_eligible_voters', 'so-berecht'),
    ('result_so_votes_total', 'so-stimmen'),
    ('result_so_turnout', 'so-bet'),
    ('result_so_votes_valid', 'so-gultig'),
    ('result_so_yeas', 'so-ja'),
    ('result_so_nays', 'so-nein'),
    ('result_so_yeas_p', 'so-japroz'),
    ('_result_so_accepted', 'so-annahme'),
    ('result_bs_eligible_voters', 'bs-berecht'),
    ('result_bs_votes_total', 'bs-stimmen'),
    ('result_bs_turnout', 'bs-bet'),
    ('result_bs_votes_valid', 'bs-gultig'),
    ('result_bs_yeas', 'bs-ja'),
    ('result_bs_nays', 'bs-nein'),
    ('result_bs_yeas_p', 'bs-japroz'),
    ('_result_bs_accepted', 'bs-annahme'),
    ('result_bl_eligible_voters', 'bl-berecht'),
    ('result_bl_votes_total', 'bl-stimmen'),
    ('result_bl_turnout', 'bl-bet'),
    ('result_bl_votes_valid', 'bl-gultig'),
    ('result_bl_yeas', 'bl-ja'),
    ('result_bl_nays', 'bl-nein'),
    ('result_bl_yeas_p', 'bl-japroz'),
    ('_result_bl_accepted', 'bl-annahme'),
    ('result_sh_eligible_voters', 'sh-berecht'),
    ('result_sh_votes_total', 'sh-stimmen'),
    ('result_sh_turnout', 'sh-bet'),
    ('result_sh_votes_valid', 'sh-gultig'),
    ('result_sh_yeas', 'sh-ja'),
    ('result_sh_nays', 'sh-nein'),
    ('result_sh_yeas_p', 'sh-japroz'),
    ('_result_sh_accepted', 'sh-annahme'),
    ('result_ar_eligible_voters', 'ar-berecht'),
    ('result_ar_votes_total', 'ar-stimmen'),
    ('result_ar_turnout', 'ar-bet'),
    ('result_ar_votes_valid', 'ar-gultig'),
    ('result_ar_yeas', 'ar-ja'),
    ('result_ar_nays', 'ar-nein'),
    ('result_ar_yeas_p', 'ar-japroz'),
    ('_result_ar_accepted', 'ar-annahme'),
    ('result_ai_eligible_voters', 'ai-berecht'),
    ('result_ai_votes_total', 'ai-stimmen'),
    ('result_ai_turnout', 'ai-bet'),
    ('result_ai_votes_valid', 'ai-gultig'),
    ('result_ai_yeas', 'ai-ja'),
    ('result_ai_nays', 'ai-nein'),
    ('result_ai_yeas_p', 'ai-japroz'),
    ('_result_ai_accepted', 'ai-annahme'),
    ('result_sg_eligible_voters', 'sg-berecht'),
    ('result_sg_votes_total', 'sg-stimmen'),
    ('result_sg_turnout', 'sg-bet'),
    ('result_sg_votes_valid', 'sg-gultig'),
    ('result_sg_yeas', 'sg-ja'),
    ('result_sg_nays', 'sg-nein'),
    ('result_sg_yeas_p', 'sg-japroz'),
    ('_result_sg_accepted', 'sg-annahme'),
    ('result_gr_eligible_voters', 'gr-berecht'),
    ('result_gr_votes_total', 'gr-stimmen'),
    ('result_gr_turnout', 'gr-bet'),
    ('result_gr_votes_valid', 'gr-gultig'),
    ('result_gr_yeas', 'gr-ja'),
    ('result_gr_nays', 'gr-nein'),
    ('result_gr_yeas_p', 'gr-japroz'),
    ('_result_gr_accepted', 'gr-annahme'),
    ('result_ag_eligible_voters', 'ag-berecht'),
    ('result_ag_votes_total', 'ag-stimmen'),
    ('result_ag_turnout', 'ag-bet'),
    ('result_ag_votes_valid', 'ag-gultig'),
    ('result_ag_yeas', 'ag-ja'),
    ('result_ag_nays', 'ag-nein'),
    ('result_ag_yeas_p', 'ag-japroz'),
    ('_result_ag_accepted', 'ag-annahme'),
    ('result_tg_eligible_voters', 'tg-berecht'),
    ('result_tg_votes_total', 'tg-stimmen'),
    ('result_tg_turnout', 'tg-bet'),
    ('result_tg_votes_valid', 'tg-gultig'),
    ('result_tg_yeas', 'tg-ja'),
    ('result_tg_nays', 'tg-nein'),
    ('result_tg_yeas_p', 'tg-japroz'),
    ('_result_tg_accepted', 'tg-annahme'),
    ('result_ti_eligible_voters', 'ti-berecht'),
    ('result_ti_votes_total', 'ti-stimmen'),
    ('result_ti_turnout', 'ti-bet'),
    ('result_ti_votes_valid', 'ti-gultig'),
    ('result_ti_yeas', 'ti-ja'),
    ('result_ti_nays', 'ti-nein'),
    ('result_ti_yeas_p', 'ti-japroz'),
    ('_result_ti_accepted', 'ti-annahme'),
    ('result_vd_eligible_voters', 'vd-berecht'),
    ('result_vd_votes_total', 'vd-stimmen'),
    ('result_vd_turnout', 'vd-bet'),
    ('result_vd_votes_valid', 'vd-gultig'),
    ('result_vd_yeas', 'vd-ja'),
    ('result_vd_nays', 'vd-nein'),
    ('result_vd_yeas_p', 'vd-japroz'),
    ('_result_vd_accepted', 'vd-annahme'),
    ('result_vs_eligible_voters', 'vs-berecht'),
    ('result_vs_votes_total', 'vs-stimmen'),
    ('result_vs_turnout', 'vs-bet'),
    ('result_vs_votes_valid', 'vs-gultig'),
    ('result_vs_yeas', 'vs-ja'),
    ('result_vs_nays', 'vs-nein'),
    ('result_vs_yeas_p', 'vs-japroz'),
    ('_result_vs_accepted', 'vs-annahme'),
    ('result_ne_eligible_voters', 'ne-berecht'),
    ('result_ne_votes_total', 'ne-stimmen'),
    ('result_ne_turnout', 'ne-bet'),
    ('result_ne_votes_valid', 'ne-gultig'),
    ('result_ne_yeas', 'ne-ja'),
    ('result_ne_nays', 'ne-nein'),
    ('result_ne_yeas_p', 'ne-japroz'),
    ('_result_ne_accepted', 'ne-annahme'),
    ('result_ge_eligible_voters', 'ge-berecht'),
    ('result_ge_votes_total', 'ge-stimmen'),
    ('result_ge_turnout', 'ge-bet'),
    ('result_ge_votes_valid', 'ge-gultig'),
    ('result_ge_yeas', 'ge-ja'),
    ('result_ge_nays', 'ge-nein'),
    ('result_ge_yeas_p', 'ge-japroz'),
    ('_result_ge_accepted', 'ge-annahme'),
    ('result_ju_eligible_voters', 'ju-berecht'),
    ('result_ju_votes_total', 'ju-stimmen'),
    ('result_ju_turnout', 'ju-bet'),
    ('result_ju_votes_valid', 'ju-gultig'),
    ('result_ju_yeas', 'ju-ja'),
    ('result_ju_nays', 'ju-nein'),
    ('result_ju_yeas_p', 'ju-japroz'),
    ('_result_ju_accepted', 'ju-annahme'),
    ('_department_in_charge', 'dep'),
    ('procedure_number', 'gesch_nr'),
    ('_position_federal_council', 'br-pos'),
    ('_position_parliament', 'bv-pos'),
    ('_position_national_council', 'nr-pos'),
    ('position_national_council_yeas', 'nrja'),
    ('position_national_council_nays', 'nrnein'),
    ('_position_council_of_states', 'sr-pos'),
    ('position_council_of_states_yeas', 'srja'),
    ('position_council_of_states_nays', 'srnein'),
    ('duration_federal_assembly', 'dauer_bv'),
    ('duration_post_federal_assembly', 'dauer_abst'),
    ('duration_initative_collection', 'i-dauer_samm'),
    ('duration_initative_federal_council', 'i-dauer_br'),
    ('duration_initative_total', 'i-dauer_tot'),
    ('duration_referendum_collection', 'fr-dauer_samm'),
    ('duration_referendum_total', 'fr-dauer_tot'),
    ('signatures_valid', 'unter_g'),
    ('signatures_invalid', 'unter_u'),
    ('_recommendation_fdp', 'p-fdp'),
    ('_recommendation_cvp', 'p-cvp'),
    ('_recommendation_sps', 'p-sps'),
    ('_recommendation_svp', 'p-svp'),
    ('_recommendation_lps', 'p-lps'),
    ('_recommendation_ldu', 'p-ldu'),
    ('_recommendation_evp', 'p-evp'),
    ('_recommendation_csp', 'p-ucsp'),
    ('_recommendation_pda', 'p-pda'),
    ('_recommendation_poch', 'p-poch'),
    ('_recommendation_gps', 'p-gps'),
    ('_recommendation_sd', 'p-sd'),
    ('_recommendation_rep', 'p-rep'),
    ('_recommendation_edu', 'p-edu'),
    ('_recommendation_fps', 'p-fps'),
    ('_recommendation_lega', 'p-lega'),
    ('_recommendation_kvp', 'p-kvp'),
    ('_recommendation_glp', 'p-glp'),
    ('_recommendation_bdp', 'p-bdp'),
    ('_recommendation_mcg', 'p-mcg'),
    ('_recommendation_sav', 'p-sav'),
    ('_recommendation_eco', 'p-eco'),
    ('_recommendation_sgv', 'p-sgv'),
    ('_recommendation_sbv_usp', 'p-sbv'),
    ('_recommendation_sgb', 'p-sgb'),
    ('_recommendation_travs', 'p-travs'),
    ('_recommendation_vsa', 'p-vsa'),
    ('_recommendation_vpod', 'p-vpod'),
    ('_recommendation_ssv', 'p-ssv'),
    ('_recommendation_gem', 'p-gem'),
    ('_recommendation_kdk', 'p-kdk'),
    ('_recommendation_vdk', 'p-vdk'),
    ('_recommendation_endk', 'p-endk'),
    ('_recommendation_fdk', 'p-fdk'),
    ('_recommendation_edk', 'p-edk'),
    ('_recommendation_gdk', 'p-gdk'),
    ('_recommendation_ldk', 'p-ldk'),
    ('_recommendation_sodk', 'p-sodk'),
    ('_recommendation_kkjpd', 'p-kkjpd'),
    ('_recommendation_bpuk', 'p-bpuk'),
    ('_recommendation_sbk', 'p-sbk'),
    ('_recommendation_acs', 'p-acs'),
    ('_recommendation_tcs', 'p-tcs'),
    ('_recommendation_vcs', 'p-vcs'),
    ('_recommendation_voev', 'p-voev'),
    ('national_council_election_year', 'nr-wahl'),
    ('national_council_share_fdp', 'w-fdp'),
    ('national_council_share_cvp', 'w-cvp'),
    ('national_council_share_sp', 'w-sp'),
    ('national_council_share_svp', 'w-svp'),
    ('national_council_share_lps', 'w-lps'),
    ('national_council_share_ldu', 'w-ldu'),
    ('national_council_share_evp', 'w-evp'),
    ('national_council_share_csp', 'w-csp'),
    ('national_council_share_pda', 'w-pda'),
    ('national_council_share_poch', 'w-poch'),
    ('national_council_share_gps', 'w-gps'),
    ('national_council_share_sd', 'w-sd'),
    ('national_council_share_rep', 'w-rep'),
    ('national_council_share_edu', 'w-edu'),
    ('national_council_share_fps', 'w-fps'),
    ('national_council_share_lega', 'w-lega'),
    ('national_council_share_kvp', 'w-kvp'),
    ('national_council_share_glp', 'w-glp'),
    ('national_council_share_bdp', 'w-bdp'),
    ('national_council_share_mcg', 'w-mcg'),
    ('national_council_share_ubrige', 'w-ubrige'),
    ('national_council_share_yeas', 'ja-lager'),
    ('national_council_share_nays', 'nein-lager'),
    ('national_council_share_neutral', 'neutral'),
    ('national_council_share_vague', 'unbestimmt'),
    ('initiator', 'urheber'),
    ('anneepolitique', 'anneepolitique'),
    ('bfs_map_de', 'bfsmap-de'),
    ('bfs_map_fr', 'bfsmap-fr'),
))


class SwissvoteDatasetField(UploadField):
    """ An upload field expecting a Swissvotes dataset (XLSX). """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('validators', [])
        kwargs['validators'].append(
            WhitelistedMimeType({
                'application/excel',
                'application/octet-stream',
                'application/vnd.ms-excel',
                'application/vnd.ms-office',
                'application/vnd.openxmlformats-officedocument.'
                'spreadsheetml.sheet',
                'application/zip'
            })
        )
        kwargs['validators'].append(FileSizeLimit(10 * 1024 * 1024))

        kwargs.setdefault('render_kw', {})
        kwargs['render_kw']['force_simple'] = True

        super().__init__(*args, **kwargs)

    def post_validate(self, form, validation_stopped):
        super(SwissvoteDatasetField, self).post_validate(
            form,
            validation_stopped
        )
        if validation_stopped:
            return

        errors = []
        data = []

        try:
            workbook = open_workbook(
                file_contents=self.raw_data[0].file.read()
            )
        except Exception:
            raise ValueError(_("Not a valid XLSX file."))

        if workbook.nsheets < 1:
            raise ValueError(_("No data."))

        sheet = workbook.sheet_by_index(0)

        if sheet.nrows <= 1:
            raise ValueError(_("No data."))

        headers = [column.value for column in sheet.row(0)]
        missing = set(COLUMNS.values()) - set(headers)
        if missing:
            raise ValueError(_(
                "Some columns are missing: ${columns}.",
                mapping={'columns': ', '.join(missing)}
            ))

        for index in range(1, sheet.nrows):
            row = sheet.row(index)
            vote = SwissVote()
            for attribute, column in COLUMNS.items():
                cell = row[headers.index(column)]
                table_column = get_table_column(vote, attribute)
                type_ = str(table_column.type)
                nullable = table_column.nullable
                try:
                    if cell.ctype == XL_CELL_EMPTY:
                        value = None
                    elif type_ == 'TEXT':
                        value = str(cell.value)
                        value = '' if value == '.' else value
                    elif type_ == 'DATE':
                        if isinstance(cell.value, str):
                            value = parse(cell.value, dayfirst=True).date()
                        else:
                            value = xldate.xldate_as_datetime(
                                cell.value,
                                workbook.datemode
                            ).date()
                    elif type_ == 'INTEGER':
                        if isinstance(cell.value, str):
                            value = cell.value
                            value = '' if value == '.' else value
                            value = int(value) if value else None
                        else:
                            value = int(cell.value)
                    elif type_ == 'INT4RANGE':
                        value = NumericRange(*[
                            int(bound) for bound in cell.value.split('-')
                        ])
                    elif type_.startswith('NUMERIC'):
                        if isinstance(cell.value, str):
                            value = cell.value
                            value = '' if value == '.' else value
                            value = Decimal(str(value)) if value else None
                        else:
                            value = Decimal(str(cell.value))
                        if value is not None:
                            prec = table_column.type.precision
                            scale = table_column.type.scale
                            value = Decimal(format(value, f'{prec}.{scale}f'))

                except Exception:
                    errors.append((
                        index, column, f"'{value}' ≠ {type_.lower()}"
                    ))

                else:
                    if not nullable and value is None:
                        errors.append((index, column, "∅"))
                    setattr(vote, attribute, value)

            data.append(vote)

        if errors:
            raise ValueError(_(
                "Some cells contain invalid values: ${errors}.",
                mapping={
                    'errors': '; '.join([
                        '{}:{} {}'.format(*error) for error in errors
                    ])
                }
            ))

        self.data = data



from webtest import TestApp as Client


def test_view_exceptions(swissvotes_app):
    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    assert (
        "Sie versuchen eine Seite zu öffnen, für die Sie nicht autorisiert "
        "sind"
    ) in client.get('/votes/update', status=403)

    assert (
        "Die angeforderte Seite konnte nicht gefunden werden."
    ) in client.get('/abstimmungen', status=404)



from cached_property import cached_property
from decimal import Decimal
from onegov.swissvotes import _


class PolicyAreaDefinition(object):
    """ Contains the static definition of a policy area. """
    def __init__(self, path=None, label=None, children=None):
        self.path = path or []
        self.decimal = None
        self.value = path[-1] if self.path else None
        self.label = label
        self.children = children or []
        self.index = {
            child.value: index for index, child in enumerate(self.children)
        }

    def get(self, key):
        """ Returns the child with the given value. """
        if key in self.index:
            return self.children[self.index[key]]


POLICY_AREA = PolicyAreaDefinition(children=[
    PolicyAreaDefinition([1], _("d-1-1"), [
        PolicyAreaDefinition([1, 11], _("d-2-11")),
        PolicyAreaDefinition([1, 12], _("d-2-12"), [
            PolicyAreaDefinition([1, 12, 121], _("d-3-121")),
            PolicyAreaDefinition([1, 12, 122], _("d-3-122")),
            PolicyAreaDefinition([1, 12, 123], _("d-3-123")),
            PolicyAreaDefinition([1, 12, 124], _("d-3-124")),
        ]),
        PolicyAreaDefinition([1, 13], _("d-2-13"), [
            PolicyAreaDefinition([1, 13, 131], _("d-3-131")),
            PolicyAreaDefinition([1, 13, 132], _("d-3-132")),
            PolicyAreaDefinition([1, 13, 133], _("d-3-133")),
            PolicyAreaDefinition([1, 13, 134], _("d-3-134")),
        ]),
        PolicyAreaDefinition([1, 14], _("d-2-14"), [
            PolicyAreaDefinition([1, 14, 141], _("d-3-141")),
            PolicyAreaDefinition([1, 14, 142], _("d-3-142")),
            PolicyAreaDefinition([1, 14, 143], _("d-3-143")),
        ]),
        PolicyAreaDefinition([1, 15], _("d-2-15"), [
            PolicyAreaDefinition([1, 15, 151], _("d-3-151")),
            PolicyAreaDefinition([1, 15, 152], _("d-3-152")),
            PolicyAreaDefinition([1, 15, 153], _("d-3-153")),
        ]),
        PolicyAreaDefinition([1, 16], _("d-2-16"), [
            PolicyAreaDefinition([1, 16, 161], _("d-3-161")),
            PolicyAreaDefinition([1, 16, 162], _("d-3-162")),
            PolicyAreaDefinition([1, 16, 163], _("d-3-163")),
            PolicyAreaDefinition([1, 16, 164], _("d-3-164")),
            PolicyAreaDefinition([1, 16, 165], _("d-3-165")),
            PolicyAreaDefinition([1, 16, 166], _("d-3-166")),
        ])
    ]),
    PolicyAreaDefinition([2], _("d-1-2"), [
        PolicyAreaDefinition([2, 21], _("d-2-21"), [
            PolicyAreaDefinition([2, 21, 211], _("d-3-211")),
            PolicyAreaDefinition([2, 21, 212], _("d-3-212")),
            PolicyAreaDefinition([2, 21, 213], _("d-3-213")),
        ]),
        PolicyAreaDefinition([2, 22], _("d-2-22"), [
            PolicyAreaDefinition([2, 22, 221], _("d-3-221")),
            PolicyAreaDefinition([2, 22, 222], _("d-3-222")),
            PolicyAreaDefinition([2, 22, 223], _("d-3-223")),
            PolicyAreaDefinition([2, 22, 224], _("d-3-224")),
        ]),
        PolicyAreaDefinition([2, 23], _("d-2-23"), [
            PolicyAreaDefinition([2, 23, 231], _("d-3-231")),
            PolicyAreaDefinition([2, 23, 232], _("d-3-232")),
        ]),
        PolicyAreaDefinition([2, 24], _("d-2-24")),
        PolicyAreaDefinition([2, 25], _("d-2-25")),
        PolicyAreaDefinition([2, 26], _("d-2-26"), [
            PolicyAreaDefinition([2, 26, 261], _("d-3-261")),
            PolicyAreaDefinition([2, 26, 262], _("d-3-262")),
        ]),
        PolicyAreaDefinition([2, 27], _("d-2-27")),
        PolicyAreaDefinition([2, 28], _("d-2-28")),
    ]),
    PolicyAreaDefinition([3], _("d-1-3"), [
        PolicyAreaDefinition([3, 31], _("d-2-31"), [
            PolicyAreaDefinition([3, 31, 311], _("d-3-311")),
            PolicyAreaDefinition([3, 31, 312], _("d-3-312")),
            PolicyAreaDefinition([3, 31, 313], _("d-3-313")),
        ]),
        PolicyAreaDefinition([3, 32], _("d-2-32"), [
            PolicyAreaDefinition([3, 32, 321], _("d-3-321")),
            PolicyAreaDefinition([3, 32, 322], _("d-3-322")),
            PolicyAreaDefinition([3, 32, 323], _("d-3-323")),
            PolicyAreaDefinition([3, 32, 324], _("d-3-324")),
            PolicyAreaDefinition([3, 32, 325], _("d-3-325")),
            PolicyAreaDefinition([3, 32, 327], _("d-3-327")),
            PolicyAreaDefinition([3, 32, 328], _("d-3-328")),
            PolicyAreaDefinition([3, 32, 329], _("d-3-329")),
        ]),
        PolicyAreaDefinition([3, 33], _("d-2-33")),
    ]),
    PolicyAreaDefinition([4], _("d-1-4"), [
        PolicyAreaDefinition([4, 41], _("d-2-41"), [
            PolicyAreaDefinition([4, 41, 411], _("d-3-411")),
            PolicyAreaDefinition([4, 41, 412], _("d-3-412")),
            PolicyAreaDefinition([4, 41, 413], _("d-3-413")),
            PolicyAreaDefinition([4, 41, 414], _("d-3-414")),
            PolicyAreaDefinition([4, 41, 415], _("d-3-415")),
            PolicyAreaDefinition([4, 41, 416], _("d-3-416")),
        ]),
        PolicyAreaDefinition([4, 42], _("d-2-42"), [
            PolicyAreaDefinition([4, 42, 421], _("d-3-421")),
            PolicyAreaDefinition([4, 42, 422], _("d-3-422")),
            PolicyAreaDefinition([4, 42, 423], _("d-3-423")),
            PolicyAreaDefinition([4, 42, 424], _("d-3-424")),
        ]),
        PolicyAreaDefinition([4, 43], _("d-2-43"), [
            PolicyAreaDefinition([4, 43, 431], _("d-3-431")),
            PolicyAreaDefinition([4, 43, 432], _("d-3-432")),
        ]),
        PolicyAreaDefinition([4, 44], _("d-2-44"), [
            PolicyAreaDefinition([4, 44, 441], _("d-3-441")),
            PolicyAreaDefinition([4, 44, 442], _("d-3-442")),
            PolicyAreaDefinition([4, 44, 443], _("d-3-443")),
        ]),
    ]),
    PolicyAreaDefinition([5], _("d-1-5"), [
        PolicyAreaDefinition([5, 51], _("d-2-51")),
        PolicyAreaDefinition([5, 52], _("d-2-52")),
        PolicyAreaDefinition([5, 53], _("d-2-53")),
        PolicyAreaDefinition([5, 54], _("d-2-54")),
        PolicyAreaDefinition([5, 55], _("d-2-55")),
    ]),
    PolicyAreaDefinition([6], _("d-1-6"), [
        PolicyAreaDefinition([6, 61], _("d-2-61"), [
            PolicyAreaDefinition([6, 61, 611], _("d-3-611")),
            PolicyAreaDefinition([6, 61, 612], _("d-3-612")),
            PolicyAreaDefinition([6, 61, 613], _("d-3-613")),
            PolicyAreaDefinition([6, 61, 614], _("d-3-614")),
        ]),
        PolicyAreaDefinition([6, 62], _("d-2-62")),
        PolicyAreaDefinition([6, 63], _("d-2-63")),
        PolicyAreaDefinition([6, 64], _("d-2-64")),
    ]),
    PolicyAreaDefinition([7], _("d-1-7"), [
        PolicyAreaDefinition([7, 71], _("d-2-71")),
        PolicyAreaDefinition([7, 72], _("d-2-72")),
        PolicyAreaDefinition([7, 73], _("d-2-73")),
        PolicyAreaDefinition([7, 74], _("d-2-74")),
        PolicyAreaDefinition([7, 75], _("d-2-75")),
    ]),
    PolicyAreaDefinition([8], _("d-1-8"), [
        PolicyAreaDefinition([8, 81], _("d-2-81"), [
            PolicyAreaDefinition([8, 81, 811], _("d-3-811")),
            PolicyAreaDefinition([8, 81, 812], _("d-3-812")),
        ]),
        PolicyAreaDefinition([8, 82], _("d-2-82"), [
            PolicyAreaDefinition([8, 82, 821], _("d-3-821")),
            PolicyAreaDefinition([8, 82, 822], _("d-3-822")),
        ]),
        PolicyAreaDefinition([8, 83], _("d-2-83"), [
            PolicyAreaDefinition([8, 83, 831], _("d-3-831")),
            PolicyAreaDefinition([8, 83, 832], _("d-3-832")),
        ]),
        PolicyAreaDefinition([8, 84], _("d-2-84")),
        PolicyAreaDefinition([8, 85], _("d-2-85")),
        PolicyAreaDefinition([8, 86], _("d-2-86")),
        PolicyAreaDefinition([8, 87], _("d-2-87")),
    ]),
    PolicyAreaDefinition([9], _("d-1-9"), [
        PolicyAreaDefinition([9, 91], _("d-2-91"), [
            PolicyAreaDefinition([9, 91, 911], _("d-3-911")),
            PolicyAreaDefinition([9, 91, 912], _("d-3-912")),
        ]),
        PolicyAreaDefinition([9, 92], _("d-2-92"), [
            PolicyAreaDefinition([9, 92, 921], _("d-3-921")),
            PolicyAreaDefinition([9, 92, 922], _("d-3-922")),
        ]),
        PolicyAreaDefinition([9, 93], _("d-2-93"), [
            PolicyAreaDefinition([9, 93, 931], _("d-3-931")),
            PolicyAreaDefinition([9, 93, 932], _("d-3-932")),
            PolicyAreaDefinition([9, 93, 933], _("d-3-933")),
            PolicyAreaDefinition([9, 93, 934], _("d-3-934")),
            PolicyAreaDefinition([9, 93, 935], _("d-3-935")),
            PolicyAreaDefinition([9, 93, 936], _("d-3-936")),
            PolicyAreaDefinition([9, 93, 937], _("d-3-937")),
            PolicyAreaDefinition([9, 93, 938], _("d-3-938")),
        ]),
    ]),
    PolicyAreaDefinition([10], _("d-1-10"), [
        PolicyAreaDefinition([10, 101], _("d-2-101"), [
            PolicyAreaDefinition([10, 101, 1011], _("d-3-1011")),
            PolicyAreaDefinition([10, 101, 1012], _("d-3-1012")),
            PolicyAreaDefinition([10, 101, 1013], _("d-3-1013")),
            PolicyAreaDefinition([10, 101, 1014], _("d-3-1014")),
            PolicyAreaDefinition([10, 101, 1015], _("d-3-1015")),
        ]),
        PolicyAreaDefinition([10, 102], _("d-2-102"), [
            PolicyAreaDefinition([10, 102, 1021], _("d-3-1021")),
            PolicyAreaDefinition([10, 102, 1022], _("d-3-1022")),
            PolicyAreaDefinition([10, 102, 1023], _("d-3-1023")),
            PolicyAreaDefinition([10, 102, 1024], _("d-3-1024")),
            PolicyAreaDefinition([10, 102, 1025], _("d-3-1025")),
            PolicyAreaDefinition([10, 102, 1026], _("d-3-1026")),
            PolicyAreaDefinition([10, 102, 1027], _("d-3-1027")),
            PolicyAreaDefinition([10, 102, 1028], _("d-3-1028")),
        ]),
        PolicyAreaDefinition([10, 103], _("d-2-103"), [
            PolicyAreaDefinition([10, 103, 1031], _("d-3-1031")),
            PolicyAreaDefinition([10, 103, 1032], _("d-3-1032")),
            PolicyAreaDefinition([10, 103, 1033], _("d-3-1033")),
            PolicyAreaDefinition([10, 103, 1034], _("d-3-1034")),
            PolicyAreaDefinition([10, 103, 1035], _("d-3-1035")),
            PolicyAreaDefinition([10, 103, 1036], _("d-3-1036")),
            PolicyAreaDefinition([10, 103, 1037], _("d-3-1037")),
            PolicyAreaDefinition([10, 103, 1038], _("d-3-1038")),
        ]),
    ]),
    PolicyAreaDefinition([11], _("d-1-11"), [
        PolicyAreaDefinition([11, 111], _("d-2-111")),
        PolicyAreaDefinition([11, 112], _("d-2-112")),
        PolicyAreaDefinition([11, 113], _("d-2-113")),
        PolicyAreaDefinition([11, 114], _("d-2-114"), [
            PolicyAreaDefinition([11, 114, 1141], _("d-3-1141")),
            PolicyAreaDefinition([11, 114, 1142], _("d-3-1142")),
        ]),
        PolicyAreaDefinition([11, 115], _("d-2-115")),
    ]),
    PolicyAreaDefinition([12], _("d-1-12"), [
        PolicyAreaDefinition([12, 121], _("d-2-121")),
        PolicyAreaDefinition([12, 122], _("d-2-122")),
        PolicyAreaDefinition([12, 123], _("d-2-123")),
        PolicyAreaDefinition([12, 124], _("d-2-124")),
        PolicyAreaDefinition([12, 125], _("d-2-125"), [
            PolicyAreaDefinition([12, 125, 1251], _("d-3-1251")),
            PolicyAreaDefinition([12, 125, 1252], _("d-3-1252")),
            PolicyAreaDefinition([12, 125, 1253], _("d-3-1253")),
            PolicyAreaDefinition([12, 125, 1254], _("d-3-1254")),
        ]),
    ])
])


class PolicyArea(object):
    """ Helper class for handling of descriptors. """

    def __init__(self, value, level=None):
        if isinstance(value, str):
            self.value = value
        elif isinstance(value, list):
            self.value = '.'.join([str(x) for x in value])
        elif isinstance(value, Decimal):
            assert level is not None
            self.value = '.'.join(
                str(int(value * 10 ** x)) for x in range(level)
            )
        else:
            raise NotImplementedError()

    def __repr__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other.value

    @cached_property
    def level(self):
        return self.value.count('.') + 1

    @cached_property
    def descriptor(self):
        return int(self.value.split('.')[-1])

    @cached_property
    def descriptor_path(self):
        return [int(part) for part in self.value.split('.')]

    @cached_property
    def descriptor_decimal(self):
        return Decimal(self.descriptor) / (10 ** (self.level - 1))

    @cached_property
    def label(self):
        return self.label_path[-1]

    @cached_property
    def label_path(self):
        result = []
        lookup = POLICY_AREA
        for part in self.descriptor_path:
            lookup = lookup.get(part)
            if not lookup:
                result.append(str(self.descriptor_decimal))
                break
            result.append(lookup.label or str(self.descriptor))
        return result

    def html(self, request):
        title = ' &gt; '.join([
            request.translate(part) for part in self.label_path
        ])
        return f'<span>{title}</span>'



from webtest import TestApp as Client


def test_view_home(swissvotes_app):
    client = Client(swissvotes_app)
    home = client.get('/').maybe_follow()
    assert "<h2>home</h2>" in home
    assert home.request.url.endswith('page/home')
